#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define prim_addr 11
#define prim_size (1<<prim_addr)
#define maxromsize (prim_size * 128)
#define init_count 64
#define init_size 32

unsigned char UCHexChar(unsigned char value);

main (int argc, char *argv[])
{
  FILE *finput, *foutput;
  unsigned long addr;
  unsigned long addr_adj;
  unsigned char data;
  unsigned char cntr;
  unsigned char init;
  unsigned char step;
  unsigned char indx_roml;
  unsigned char indx_romh;
  unsigned char indx_muxl;
  unsigned char indx_muxh;
  unsigned char indx_addr;
  unsigned char romcount;
  unsigned char muxrequired;
  unsigned char rom[maxromsize];

  // If we don't get the right number of input arguments,
  // exit and return some helpful information to the user.

  if (argc != 4)
    {
      printf ("Usage:    raw2ver <binary input file> <verilog output file> <module name>\n\n");
	  printf ("Function: Writes a template for a byte-wide rom using unisim\n");
	  printf ("          library blockram instances.  The maximum rom size is\n");
	  printf ("          limited to 128 blockrams and is built using ramb16_s9.\n");
      return (1);
    }

  // If we do get the right number of input arguments,
  // check that the binary input file can be opened for
  // reading and that the output text file can be opened
  // for writing.

  finput = fopen (argv[1], "rb");
  if (finput == NULL)
    {
      printf ("Error opening '%s' for reading.\n", argv[1]);
      return (2);
    }

  foutput = fopen (argv[2], "w");
  if (foutput == NULL)
    {
      printf ("Error opening '%s' for writing.\n", argv[2]);
      fclose (finput);
      return (3);
    }

  // Read the file into the rom array.  This allows us to
  // determine the rom size and later access the data easily
  // when writing out the unisim primitive initializations.
  // If for some weird reason, we get a zero length file,
  // just exit with an informational note.

  for (addr = 0; addr < maxromsize ; addr++) rom[addr] = 0;
  addr = 0;
  fread (&data, sizeof(data), 1, finput);
  while (!feof(finput) && (addr < maxromsize))
    {
      rom[addr] = data;
	  addr++;
	  fread (&data, sizeof(data), 1, finput);
    }
  printf ("Binary input is %d bytes.\n", addr);
  fclose (finput);
  if (addr == 0)
    {
      printf ("Error due to zero-byte file.\n");
      fclose (finput);
	  fclose (foutput);
      return (4);
    }

  // Start off assuming the data fits in a single primitive.
  // the primitive address indices are known constants.

  addr_adj = prim_size;
  indx_roml = 0;
  indx_romh = prim_addr - 1;

  // Search to find out if multiple primitives are needed.
  // Double the rom size until we find a size that fits all
  // the data.  Keep track of what the high index would be.

  indx_muxl = prim_addr;
  indx_muxh = prim_addr - 1;
  while (addr_adj < addr)
    {
	  addr_adj = addr_adj << 1;
	  indx_muxh++;
	}
  indx_addr = indx_muxh;

  // Determine how many primitive instances are required.

  romcount = (unsigned char)(((addr - 1) >> prim_addr) + 1);
  addr = romcount * prim_size;

  // Determine if we will need to build a multiplexer to
  // mux the outputs of multiple primitive instances.

  if (indx_muxh < indx_muxl) muxrequired = 0;
  else muxrequired = 1;

  // Print out some diagnostic information to the console.

  printf ("ROM lo index: %d\n", indx_roml);
  printf ("ROM hi index: %d\n", indx_romh);
  if (muxrequired)
	{
      printf ("MUX lo index: %d\n", indx_muxl);
      printf ("MUX hi index: %d\n", indx_muxh);
	}
  else printf ("MUX not required.\n");
  printf ("Address index: %d\n", indx_addr);
  printf ("ROM size is %d bytes.\n", addr);
  printf ("Instance count is %d.\n", romcount);
  
  // Write out the header information for this verilog rom.
  // This includes the module and port names.

  fprintf (foutput, "/***********************************************************************\n");
  fprintf (foutput, "\n");
  fprintf (foutput, "  File: %s\n", argv[2]);
  fprintf (foutput, "\n");
  fprintf (foutput, "  This is a BlockROM generated by the raw2ver program.\n");
  fprintf (foutput, "\n");
  fprintf (foutput, "  Copyright (c) 2005 Eric Crabill.  All rights reserved.  This design\n");
  fprintf (foutput, "  is provided solely for non-commercial, educational purposes.  Eric\n");
  fprintf (foutput, "  Crabill makes no warranties, and accepts no liability, with respect\n");
  fprintf (foutput, "  to this design or its use, and any use thereof is solely at the risk\n");
  fprintf (foutput, "  of the user.  This copyright notice must be retained as part of this\n");
  fprintf (foutput, "  text at all times.\n");
  fprintf (foutput, "\n");
  fprintf (foutput, "***********************************************************************/\n");
  fprintf (foutput, "\n");
  fprintf (foutput, "`timescale 1 ns / 1 ps\n");
  fprintf (foutput, "\n");
  fprintf (foutput, "module %s (addr, dout, clk, rst);\n", argv[3]);
  fprintf (foutput, "\n");
  fprintf (foutput, "  input  [%d:0] addr;\n", indx_addr);
  fprintf (foutput, "  output  [7:0] dout;\n");
  fprintf (foutput, "  input         clk;\n");
  fprintf (foutput, "  input         rst;\n");
  
  // This is the meat of the program.  Loop over the number
  // of instances required.  For each instance, three things
  // need to be done.  First, the instance and its output
  // net need to be created.  Second, the synthesis attributes
  // need to be created based on the data we read from the
  // file.  Third, the defparams for simulation need to be
  // added for proper simulation.

  fprintf (foutput, "\n");
  fprintf (foutput, "\n");
  fprintf (foutput, "  //******************************************************************//\n");
  fprintf (foutput, "  // The following section implements the BlockROM instance(s).       //\n");
  fprintf (foutput, "  //******************************************************************//\n");
  fprintf (foutput, "\n");

  for (cntr = 0; cntr < romcount; cntr++)
	{
      fprintf (foutput, "\n");
	  fprintf (foutput, "  wire    [7:0] rom%d_out;\n", cntr);
	  fprintf (foutput, "\n");
	  fprintf (foutput, "  RAMB16_S9 ROM%d (\n", cntr);
	  fprintf (foutput, "                .DO(rom%d_out), .DOP(),\n", cntr);
	  fprintf (foutput, "                .ADDR(addr[%d:%d]),\n", indx_romh, indx_roml);
	  fprintf (foutput, "                .DI(8'h00), .DIP(1'b0),\n");
	  fprintf (foutput, "                .EN(1'b1), .WE(1'b0),\n");
	  fprintf (foutput, "                .SSR(rst), .CLK(clk));\n");
	  // This is for the synthesis attributes.
      fprintf (foutput, "\n");
	  for (init = 0; init < init_count; init++)
		{
		  fprintf (foutput, "  // synthesis attribute INIT_");
		  data = (init >> 4);
		  data = UCHexChar(data);
		  fprintf (foutput, "%c", data);
		  data = (init & 0x0f);
		  data = UCHexChar(data);
		  fprintf (foutput, "%c of ROM%d is \"", data, cntr);
		  addr_adj = (prim_size * cntr) + (init_size * init);
		  for (step = 0; step < init_size; step++)
			{
			  addr = addr_adj + init_size - step - 1;
			  data = rom[addr];
			  data = (data >> 4);
			  fprintf (foutput, "%1x", data);
			  data = rom[addr];
			  data = (data & 0x0f);
			  fprintf (foutput, "%1x", data);
			}
		  fprintf (foutput, "\"\n");
		}
	  // This is for the simulation attributes.
	  fprintf (foutput, "\n");
	  fprintf (foutput, "  // synthesis translate_off\n");
	  for (init = 0; init < init_count; init++)
		{
		  fprintf (foutput, "  defparam ROM%d.INIT_", cntr);
		  data = (init >> 4);
		  data = UCHexChar(data);
		  fprintf (foutput, "%c", data);
		  data = (init & 0x0f);
		  data = UCHexChar(data);
		  addr_adj = init_size * 8;
		  fprintf (foutput, "%c = %d'h", data, addr_adj);
		  addr_adj = (prim_size * cntr) + (init_size * init);
		  for (step = 0; step < init_size; step++)
			{
			  addr = addr_adj + init_size - step - 1;
			  data = rom[addr];
			  data = (data >> 4);
			  fprintf (foutput, "%1x", data);
			  data = rom[addr];
			  data = (data & 0x0f);
			  fprintf (foutput, "%1x", data);
			}
		  fprintf (foutput, ";\n");
		}
	  fprintf (foutput, "  // synthesis translate_on\n");
	}

  // If an output multiplexer is required, implement it here.
  // If an output multiplexer is not required, simply pass out
  // the single memory output.

  if (muxrequired == 1)
	{
	  fprintf (foutput, "\n");
      fprintf (foutput, "\n");
      fprintf (foutput, "  //******************************************************************//\n");
      fprintf (foutput, "  // The following section implements the output multiplexer.         //\n");
      fprintf (foutput, "  //******************************************************************//\n");
      fprintf (foutput, "\n");
      fprintf (foutput, "\n"); 
      fprintf (foutput, "  reg     [7:0] reg_dout;\n");
      fprintf (foutput, "\n"); 
      fprintf (foutput, "  always @(\n");
	  for (cntr = 0; cntr < romcount; cntr++)
		  fprintf (foutput, "          rom%d_out or\n", cntr);
      fprintf (foutput, "          addr\n");
	  fprintf (foutput, "          )\n");
      fprintf (foutput, "  begin\n");
	  fprintf (foutput, "    case (addr[%d:%d])\n", indx_muxh, indx_muxl);
	  for (cntr = 0; cntr < romcount; cntr++)
		  fprintf (foutput, "      %d: reg_dout = rom%d_out;\n", cntr, cntr);
      fprintf (foutput, "      default: reg_dout = 8'h00;\n");
      fprintf (foutput, "    endcase\n");
      fprintf (foutput, "  end\n");
      fprintf (foutput, "\n"); 
      fprintf (foutput, "  assign dout = reg_dout;\n");
	}
  else
	{
	  fprintf (foutput, "\n");
      fprintf (foutput, "\n");
      fprintf (foutput, "  //******************************************************************//\n");
      fprintf (foutput, "  // The following section implements the output assignment.          //\n");
      fprintf (foutput, "  //******************************************************************//\n");
      fprintf (foutput, "\n");
      fprintf (foutput, "\n"); 
      fprintf (foutput, "  assign dout = rom0_out;\n");
	}

  // Write out the footer information for this verilog rom.
  // This should wrap up the file generation process, so it
  // is okay to close the output file.

  fprintf (foutput, "\n");
  fprintf (foutput, "\n");
  fprintf (foutput, "  //******************************************************************//\n");
  fprintf (foutput, "  //                                                                  //\n");
  fprintf (foutput, "  //******************************************************************//\n");
  fprintf (foutput, "\n");
  fprintf (foutput, "\n");
  fprintf (foutput, "endmodule\n");
  fclose (foutput);
  return (0);
}

unsigned char UCHexChar(unsigned char value)
{
  unsigned char result;

  switch (value)
	{
	  case 0:
		result = '0';
	    break;
	  case 1:
	    result = '1';
	    break;
      case 2:
	    result = '2';
	    break;
      case 3:
		result = '3';
		break;
	  case 4:
		result = '4';
	    break;
	  case 5:
	    result = '5';
	    break;
      case 6:
	    result = '6';
	    break;
      case 7:
		result = '7';
		break;
	  case 8:
		result = '8';
	    break;
	  case 9:
	    result = '9';
	    break;
      case 10:
	    result = 'A';
	    break;
      case 11:
		result = 'B';
		break;
	  case 12:
		result = 'C';
	    break;
	  case 13:
	    result = 'D';
	    break;
      case 14:
	    result = 'E';
	    break;
      case 15:
		result = 'F';
		break;
	  default:
	    result = 'X';
	    break;
	}
  return(result);
}
